#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "global.h"
#include "err.h"

/**********************************************************************/

SCS_inst*  new_scs_instance
(
  long  alpha_size,
  long  str_len,
  long  num_str
)
{
  long       i;
  SCS_inst*  scs_inst;
  Alpha**    str;
  Alpha*     str_pool; 

  scs_inst = (SCS_inst*)calloc(1, sizeof(SCS_inst) );
  str      = (Alpha**)calloc( num_str, sizeof(Alpha*) );
  str_pool = (Alpha*)calloc( num_str*str_len, sizeof(Alpha) );

  if( !scs_inst || !str || !str_pool )
  {
    err_exit( "Cannot allocate memory in `new_scs_instance'!" );
  }

  scs_inst->alpha_size = alpha_size;
  scs_inst->num_str    = num_str;
  scs_inst->str_len    = str_len;
  scs_inst->str_pool   = str_pool;
  for( i = 0;   i < num_str;  i++ )
  {
    str[i] = &( str_pool[i*str_len] );
  }
  scs_inst->str = str;
  scs_inst->hidden_super_seq_len = 0;
  scs_inst->hidden_super_seq = (Alpha*)NULL;

  return  scs_inst;
}

/**********************************************************************/

void  free_scs_instance
(
  SCS_inst*  scs_inst
)
{
  if( !scs_inst)  return;
  if( scs_inst->str_pool )
  {
    free( scs_inst->str_pool );
  }
  if( scs_inst->hidden_super_seq )
  {
    free( scs_inst->hidden_super_seq );
  }
  free( scs_inst );
}

/**********************************************************************/
void  print_string
( 
  long    str_len,
  Alpha*  str
)
{
  long  i;
  for( i = 0;  i < str_len;  i++ ) 
  {
    alpha_out( str[i] );
  }
  printf( "\n" );   
}

/**********************************************************************/

void  print_scs_instance
( 
  SCS_inst*  scs_inst
)
{
  long    i;

  assert( scs_inst );

  printf( "ALPHA_SIZE %6ld\n", scs_inst->alpha_size );
  printf( "STR_LEN %6ld\n",    scs_inst->str_len );
  printf( "NUM_STR %6ld\n",    scs_inst->num_str );
 
  for( i = 0;  i < scs_inst->num_str;  i++ )
  {
    print_string( scs_inst->str_len, scs_inst->str[i] );
  }

  if( scs_inst->hidden_super_seq )
  {
    printf( "HIDDEN_SUPER_SEQ_LEN: %ld\n", scs_inst->hidden_super_seq_len );
    printf( "HIDDEN_SUPER_SEQ: " );
    print_string( scs_inst->hidden_super_seq_len, 
                  scs_inst->hidden_super_seq );
  } 
}

/**********************************************************************/
SCS_inst*  read_scs_instance()
{
  long       i, j;
  long  alpha_size;
  long  str_len;
  long  num_str;
  Alpha*     str;
  SCS_inst*  scs_inst;

  scanf( "ALPHA_SIZE %ld\n", &alpha_size );
  scanf( "STR_LEN %ld\n",    &str_len );
  scanf( "NUM_STR %ld\n",    &num_str );

  scs_inst = new_scs_instance( alpha_size, str_len, num_str );
  if( scs_inst ) 
  {
    for( i = 0;  i < num_str;  i++ )
    {
      str = scs_inst->str[i];
      for( j = 0;  j < str_len;  j++ ) 
      {
        alpha_in( str[j] );
        assert( (0 <= str[j]) && (str[j] < alpha_size) );
      }
      scanf( "\n" );
    }
  }

  if( scanf( "HIDDEN_SUPER_SEQ_LEN: %ld\n", &(scs_inst->hidden_super_seq_len) ) == 1 )
  {
    str = scs_inst->hidden_super_seq = 
      (Alpha*)calloc( scs_inst->hidden_super_seq_len, sizeof(Alpha) );
    if( !(scs_inst->hidden_super_seq) )
    {
      err_exit( "Cannot allocate memory in `read_scs_instance'!" );
    }
    scanf( "HIDDEN_SUPER_SEQ: " );
    for( j = 0;  j < scs_inst->hidden_super_seq_len;  j++ ) 
    {
      alpha_in( str[j] );
    }
    scanf( "\n" );   
  } 

  return  scs_inst;
}

/**********************************************************************/

void  print_tight_str
(
  SCS_inst*  scs_inst,
  long       super_seq_len,
  Alpha*     super_seq
)
{
  long    i, j, k, num_tight, span;
  Alpha*  str;

  span = 0;
  for( i = 0;  i < scs_inst->num_str;  i++ )
  {  
    str = scs_inst->str[i];
    k = 0;
    for( j = 0;  j < scs_inst->str_len;  j++ )
    {
      while( (k < super_seq_len) && (str[j] != super_seq[k]) )
      {
        k++;
      }
      k++;
      if( k > super_seq_len )
      {
        break;
      }
    }
    if( j < scs_inst->str_len )
    {
      return;
    }
    else
    {
      span = MAX(k, span);
    }
  }

  printf( "BEGIN_TIGHT_STRING_LIST\n" );
  num_tight = 0;

  for( i = 0;  i < scs_inst->num_str;  i++ )
  {
    str = scs_inst->str[i];
    k = 0;
    for( j = 0;  j < scs_inst->str_len;  j++ )
    {
      while( (k < super_seq_len) && (str[j] != super_seq[k]) )
      {
        k++;
      }
      k++;
      if( k > super_seq_len )
      {
        break;
      }
    }
    if( k > span - 4 )
    {
      num_tight++;
      print_string(  scs_inst->str_len, str );
    }
  }

  printf( "END_TIGHT_STRING_LIST\n" );
  printf( "NUM_TIGHT_STRINGS: %ld\n", num_tight );
}

/*********************************************************************** 
***********************************************************************/
