#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <ctype.h>

#define TRUE 1
#define FALSE 0
#define max_m 251
#define max_n 1001

#undef _DEBUG

static char code[5] = {' ', 'a', 'c', 't', 'g'};
static char decode[2][256];

struct pair_struct
{
  char str[2][max_n];
  long pid[2][max_n];
  long n[2];
  long start[2];
};
typedef struct pair_struct PairStruct;

struct primer_struct
{
  long pair, str, pos, deg, aux;
};
typedef struct primer_struct PrimerStruct;

long prox =0;

inline long find_primer(
  PairStruct*  pair,
  long  i,
  long  str,
  long  pos, 
  PrimerStruct* primer, 
  long num_primers,
  long k
)
{
  long  p, j;
  long  found;
  
  for( p = 0;  p < num_primers;  p++ )
  {
    found = TRUE;
    for( j = 0; j < k; j++ )
    {
      if( pair[i].str[str][pos+j] != 
	  pair[primer[p].pair].str[primer[p].str][primer[p].pos+j])
      {
        found = FALSE;
        break;
      }
    }
    if( found ) return p;
  }

   return -1;
}

inline long first_primer(
  PairStruct*  pair,
  long  i,
  long  str,
  long  pos, 
  long k
)
{
  long  j, pos1;
  long  found;
  
 //XXX for( pos1 = pair[i].start[str]; pos1 < pos;  pos1++ )
  for(pos1=pos+1; pos1 < pair[i].n[str]-prox-k; pos1++)
  {
    found = TRUE;
    for( j = 0; j < k; j++ )
    {
      if( pair[i].str[str][pos+j] != pair[i].str[str][pos1+j] )
      {
        found = FALSE;
        break;
      }
    }
    if( found ) return FALSE;
  }

   return TRUE;
}


int main(int argc, char *argv[])
{
  char c;
  long i, j, pid, str, pos, num_p, m, p;
 // long num_primers;
  long num_unique;
  long k = 8;  /* default primer length*/
  long L = 1000;  /* default product length*/
  prox = 0;   /* dafault min distance between primer and amplification locus*/
  long  unrecognized_param = FALSE;
  PairStruct  pair[max_m];
  PrimerStruct  primer[max_n*max_m];  

  char  *output=NULL;
  short INTOFILE = FALSE;
  short hide_glpsol_msg = FALSE;
  FILE *fp,*fr;



/*  BEGIN - DEFINITIONS FOR PRIMER SPECIFICATION   */
#define MAX_CANDID_TEMP      2000.0  /* max melting temp of the canddate */
#define MIN_CANDID_TEMP     -273.0  /* max melting temp of the canddate */
#define DEFAULT_CANDID_SALT_CONC     50  /* is 50.0 milimolar */
#define DEFAULT_CANDID_DNA_CONC      50  /* is 50.0 nanomolar */
#define MAX_NN_TM_LENGTH 36 /* The maxium length for which to use the
                               nearest neighbor model when calculating
                               oligo Tms. */
#define MAX_CANDID_GC     100  /* max gc content of the distinguisher */
#define MIN_CANDID_GC     0  /* max gc content of the distinguisher */

/*  END - DEFINITIONS FOR PRIMER SPECIFICATION   */


/*  BEGIN -  VARIABLES AND FUNCTIONS FOR PRIMER PROPERTIES */
   double temperature, gc_content;               
   double gc_and_n_content(const char *, int );
   double seqtm(const char *,  /* The sequence. */
             double,   /* DNA concentration (nanomolar). */
             double,  /* Salt concentration (millimolar). */
             int, /* The maximum sequence length for using the
                                   nearest neighbor model (as implemented
                                   in oligotm. For sequences longer
                                   than this use the "GC%" formula implemented
                                   in long_seq_tm. */
             int       /* this is the length of the sequence */
             );


   double min_candid_temp = MIN_CANDID_TEMP;    /* min melting temp of distinguisher */
   double max_candid_temp = MAX_CANDID_TEMP;    /* max melting temp of distinguisher */
   double min_candid_gc = MIN_CANDID_GC;    /* min  gc content of distinguisher */
   double max_candid_gc = MAX_CANDID_GC;    /* max  gc content of distinguisher */
   double candid_salt_conc = DEFAULT_CANDID_SALT_CONC;
   double candid_dna_conc = DEFAULT_CANDID_SALT_CONC;
/*  END -  VARIABLES  AND FUNCTIONS FOR PRIMER PROPERTIES */

      for(i =1; i<argc; i++)
       {
        if( strncmp(argv[i], "-hide_glpsol_msg", strlen("-hide_glpsol_msg")) == 0 )
         {
          hide_glpsol_msg=TRUE;
         }
        else if( strncmp(argv[i], "-output", strlen("-output")) == 0 )
         {
          output= argv[++i];
          INTOFILE=TRUE;
         }
        else if( strncmp(argv[i], "-prodlen", strlen("-prodlen")) == 0 )
         {
          L= atof(argv[++i]);
         }
        else if( strncmp(argv[i], "-prox", strlen("-prox")) == 0 )
         {
          prox= atol(argv[++i]);
         }
        else if( strncmp(argv[i], "-plen", strlen("-plen")) == 0 )
         {
          k= atol(argv[++i]);
         }
        else if( strncmp(argv[i], "-max_melt_temp", strlen("-max_melt_temp")) == 0 )
         {
          max_candid_temp = atof(argv[++i]);
         }
        else if( strncmp(argv[i], "-min_melt_temp", strlen("-min_melt_temp")) == 0 )
         {
          min_candid_temp = atof(argv[++i]);
         }
        else if( strncmp(argv[i], "-min_gc", strlen("-min_gc")) == 0 )
         {
          min_candid_gc = atof(argv[++i]);
         }
        else if( strncmp(argv[i], "-max_gc", strlen("-max_gc")) == 0 )
         {
          max_candid_gc = atof(argv[++i]);
         }
        else if( strncmp(argv[i], "-salt_conc", strlen("-salt_conc")) == 0 )
         {
          candid_salt_conc = atof(argv[++i]);
         }
        else if( strncmp(argv[i], "-dna_conc", strlen("-dna_conc")) == 0 )
         {
          candid_dna_conc = atof(argv[++i]);
         }
        else
          {
            unrecognized_param = TRUE;
          }
        }

      if(unrecognized_param)
      {
         fprintf(stderr,"Usage:\n \"%s [OPTIONS] < filename\" or \"cat filename | %s\" \n\n", argv[0],argv[0]);
         fprintf(stderr, "OPTIONS\n"); 
         fprintf(stderr,"-plen N              : primer length (default: 8)\n");
         fprintf(stderr,"-prox N              : minimum distance between primer and amplification locus (default: 0)\n");
         fprintf(stderr,"-prodlen N           : maximum product length (default: 100)\n");
         fprintf(stderr,"-output file         : output file name (default: standard output)\n");
         fprintf(stderr,"-max_melt_temp N          : maximum primer temperature (default: none)\n");
         fprintf(stderr,"-min_melt_temp N          : maximum primer temperature (default: none)\n");
         fprintf(stderr,"-min_gc N            : minimum gc content of the primers (default: 0 %%)\n");
         fprintf(stderr,"-max_gc N            : maximum gc content of the primers (default: 100 %%)\n");
         fprintf(stderr,"-salt_conc N         : salt concentration in the reaction (default: 50 milimolar)\n" );
         fprintf(stderr,"-dna_conc N          : dna concentration in the reaction (default: 50 nanomolar)\n" );
         exit(0);
       }

if(INTOFILE==TRUE)
   {
    if((fr=fopen(output,"w"))==0)
     {
       printf("Error: Cannot open output file\n"); 
       exit(0);
     }
   }
  else 
    fr=stdout;


  // initialize decoding table
  for( j = 0; j < 2;  j++ ) 
    for( i = 0; i < 256;  i++ ) 
    { 
      decode[j][i] = 0;
    }
  // input strings are assumed to be in 5'-3' order
  // first second is complemented, then reversed 
  decode[0][(int)'a'] = decode[0][(int)'A'] = 
  decode[1][(int)'t'] = decode[1][(int)'T'] = 1;
  decode[0][(int)'c'] = decode[0][(int)'C'] = 
  decode[1][(int)'g'] = decode[1][(int)'G'] = 2;
  decode[0][(int)'t'] = decode[0][(int)'T'] = 
  decode[1][(int)'a'] = decode[1][(int)'A'] = 3;
  decode[0][(int)'g'] = decode[0][(int)'G'] = 
  decode[1][(int)'c'] = decode[1][(int)'C'] = 4;

  m = 0;
  while( !feof(stdin))
  {
    // read pair
    while( ((c=getchar()) != '<') && !feof(stdin) );

    /* first string */
    str = 0; pos = 0;  
    while( ((c=getchar()) != ',') && !feof(stdin) ) 
    { 
      if( decode[str][(int)c] > 0) 
      {
        pair[m].str[str][pos] = decode[str][(int)c];
        pos++; 
        if( pos > max_n ) 
        {
          printf( "Input string longer than %d\n", max_n );
          return -1;
        }
      }
    }
    pair[m].n[str] = pos; 
    if( !feof(stdin) )  c=getchar(); // skip over ','


    /* second string */
    str = 1;  pos = 0;  
    while( ((c=getchar()) != '>') && !feof(stdin) )  
    {
      if( decode[str][(int)c] > 0)
      { 
        pair[m].str[str][pos] = decode[str][(int)c];
        pos++; 
        if( pos > max_n ) 
        {
          printf( "Input string longer than %d\n", max_n );
          return -1;
        }
      }
    }
    pair[m].n[str] = pos; 

  // Truncate strings if needed
    for( str = 0; str < 2; str++ )
    {
      pair[m].start[str] = 0;
      if(pair[m].n[str] > L)
      {
    //XXX this if block has to be added for the left string 
        if(str==0)
         {
          for(i = 0;  i < L;  i++ )
            pair[m].str[str][i] = pair[m].str[str][i+ pair[m].n[str]-L];
         }
        pair[m].n[str] = L;
      }
    }


  /* reverse second  string */
    str =1;
    for(i = 0;  i < pair[m].n[str]/2;  i++ )
    {
       c  = pair[m].str[str][i]; 
       pair[m].str[str][i] = pair[m].str[str][pair[m].n[str] - i-1];
       pair[m].str[str][pair[m].n[str] - i-1] = c; 
    }



#ifdef _DEBUG

    for( i = 0;  i < pair[m].n[0];  i++ )
     {
       printf("%c",code[(int)pair[m].str[0][i]]); 
     }

    printf(" ");

    for( i = 0;  i < pair[m].n[1];  i++ )
     {
       printf("%c",code[(int)pair[m].str[1][i]]); 
     }

   printf("\n");
#endif


   if( !feof(stdin) )  
    {
      c=getchar();  // skip over '>'
      m++;
    }
  }

  num_p = 0;
  for( i = 0;  i < m;  i++ )
  {  
    for( str = 0; str < 2; str++ )
    {
   //XXX   for( pos = pair[i].start[str];  pos < (pair[i].n[str]- prox - k + 1);  pos++ )
      for(pos=(pair[i].n[str]-prox-k+1)-1; pos >= pair[i].start[str]; pos--)
      {
        pid = find_primer(pair, i, str, pos, primer, num_p, k); 
        if( pid < 0 ) // new primer found, add to list
        {
          primer[num_p].pair = i;
          primer[num_p].str  = str;
          primer[num_p].pos  = pos;
          primer[num_p].deg  = 1;
          pair[i].pid[str][pos] = num_p;
          num_p++;
        }  // primer already discovered
        else if(first_primer(pair, i, str, pos, k) )  // first occurence in this string
        {
          pair[i].pid[str][pos] = pid;
          (primer[pid].deg)++;    
        }
        else  // primer appears earlier in the string; mark as deleted
        {
          pair[i].pid[str][pos] = -1;
        }
      }

      for( pos = (pair[i].n[str]-prox - k + 1); pos < pair[i].n[str];  pos++ )
      {
        pair[i].pid[str][pos] = -1;
      }
    }
  }


   char *Seq, *Sequen;
    
   if( (Seq =(char *)malloc( (k+1)*sizeof(char)))==NULL)
     {
        printf("Cannot allocate memory for a short string\n");
        exit(0);
     }

   if( (Sequen =(char *)malloc( (k+1)*sizeof(char)))==NULL)
     {
        printf("Cannot allocate memory for a short string\n");
        exit(0);
     }


/*  remove the candidate primes that do not satisfy the bio-chemical constraints*/
    for( p = 0;  p < num_p;  p++ )
    {
        for( j = 0; j < k; j++ )
        {
           Seq[j]=toupper(code[ (int)pair[primer[p].pair].str[primer[p].str][primer[p].pos+j] ] );
           Sequen[j]=code[(int)pair[primer[p].pair].str[primer[p].str][primer[p].pos+j]];

        }                  
        Seq[k] ='\0';
        Sequen[k] ='\0';

        temperature=seqtm(Seq, candid_salt_conc, candid_dna_conc, MAX_NN_TM_LENGTH, strlen(Seq));

        gc_content=gc_and_n_content(Seq, strlen(Seq));

        if((temperature<min_candid_temp)||(temperature>max_candid_temp)\
             ||(gc_content<min_candid_gc)||(gc_content>max_candid_gc))
        {
           primer[p].deg = 0;
        }

    }


  // mark the pid positions on the strings -1 if the primer has deg =0
  for( i = 0;  i < m;  i++ )
    for( str = 0; str < 2; str++ )
      for( pos = pair[i].start[str];  pos < (pair[i].n[str] - k + 1);  pos++ )
      {
        pid = pair[i].pid[str][pos];
        if(primer[pid].deg==0)
         {
           pair[i].pid[str][pos]=-1;
         }
      }

  /* remove dominated unique primers  */

  for( i = 0;  i < m;  i++ )
  {  
    for( str = 0; str < 2; str++ )
    {
      num_unique = 0;
      for( pos = pair[i].n[str] - k ; pos >= pair[i].start[str]; pos-- )
      {
        pid = pair[i].pid[str][pos];
        if( pid < 0 ) continue;

        if( primer[ pid ].deg == 1 )
        {
          num_unique++;
          if( num_unique > 1 ) { pair[i].pid[str][pos] = -1; primer[ pid ].deg = 0;}
        }

        int temp;
        temp = find_primer(pair, i, str, pos, primer, num_p, k); 
         if(temp!=pid) perror("assert violation\n");
        if(primer[pid].deg ==0)
           pair[i].pid[str][pos] = -1; 

      }
    }
  }

    if((fp=fopen(".ILP-FILE","w"))==0)    
      {
         printf("Cannot open file to write the ILP\n") ;
         exit(0);
      }

  /* Writing the ILP formulation */
    fprintf(fp,"\nMinimize\n\t x(0)" );
    for( p = 1; p < num_p;  p++ ) 
      if(primer[p].deg>0) 
       { 
         fprintf(fp, " + x(%ld)", p ); 
       }
    fprintf(fp, "\n" );

#ifdef _DEBUG   
    printf("Primer        Seq pair   L/R(0/1)   Pos   Degree\n");
    printf("------        --------   --------   ---   ------\n");
#endif

    for( p = 0;  p < num_p;  p++ )
    {
      for( j = 0; j < k; j++ )
        Sequen[j]=code[(int)pair[primer[p].pair].str[primer[p].str][primer[p].pos+j]];
      Sequen[k] ='\0';


#ifdef _DEBUG   
      if(primer[p].deg != 0)
        printf(" %-15s   %ld       %ld        %ld        %ld\n", Sequen,primer[p].pair, \
          primer[p].str, primer[p].pos, primer[p].deg);
#endif

    }

    fprintf(fp,"Subject To\n");

    for( i = 0;  i < m;  i++ )
    {  
      for( str = 0;  str < 2;  str++ )
      {  

/* print the first end linear inequality */
	pos = pair[i].n[str] -prox-1 -k+1;        
        pid = pair[i].pid[str][pos];
        if(pid >= 0)
            fprintf(fp, "\t x(%ld)  -covered(%ld,%ld,%ld) >= 0\n", pid, i, str, pos );
        else
            fprintf(fp, "\t  -covered(%ld,%ld,%ld) >= 0\n", i, str, pos );

       for( j = 0; j < k; j++ )
         Sequen[j]=code[(int)pair[i].str[str][pos+j]];
       Sequen[k] ='\0';


/* print the rest of the linear inequalities */
     //XXX   for( pos = pair[i].start[str];  pos < pair[i].n[str]-1 -k+1;  pos++)
        for( pos = pair[i].n[str]-prox-1 -k+1 -1;  pos >= pair[i].start[str]; pos--)
        {
	  pid = pair[i].pid[str][pos];
          fprintf(fp, "\t covered(%ld,%ld,%ld) ", i, str, pos+1 );
          if( pid >= 0 )
	    fprintf(fp, " + x(%ld)", pid );



          for( j = 0; j < k; j++ )
            Sequen[j]=code[(int)pair[i].str[str][pos+j]];
          Sequen[k] ='\0';

          fprintf(fp, " - covered(%ld,%ld,%ld) >= 0\n", i, str, pos );
        }
      }
    }


    for( i = 0;  i < m;  i++ )
    {  
    /* the contraints that imposes product length*/
       fprintf(fp, "\t " );
        for( str = 0;  str < 2;  str++ )
          {  
            for( pos = pair[i].start[str];  pos <= pair[i].n[str]-prox-1-k+1;  pos++)
             {
               fprintf(fp, "+ covered(%ld,%ld,%ld)", i, str, pos );
             }
          }
          fprintf(fp," >= %ld\n", pair[i].n[0] + pair[i].n[1]+2- L);


     /* forces both of the strings of each pair to be covered*/
         for( str = 0;  str < 2;  str++ )
          {  
            fprintf(fp, "\t " );
            for( pos = pair[i].start[str];  pos <= pair[i].n[str]-prox-1-k+1;  pos++)
             {
               fprintf(fp, "+ covered(%ld,%ld,%ld)", i, str, pos );
             }
            fprintf(fp," >= %d\n",1);
          }
    }


    fprintf(fp,"Binaries\n");
    for( p = 0; p < num_p;  p++ ) if(primer[p].deg >0) fprintf(fp, "\t x(%ld)\n", p );

    for( i = 0;  i < m;  i++ )
    {  
      for( str = 0;  str < 2;  str++ )
      {  
        for( pos = pair[i].start[str];  pos < pair[i].n[str]-prox-k+1;  pos++)
        {
          fprintf(fp, "\t covered(%ld,%ld,%ld)\n", i, str, pos );
        }
      }
    }

    fprintf(fp,"End\n"); 
    fclose(fp);

    /* Solve the Integer Program using GLPK SOLVER */
    if( hide_glpsol_msg ) system("glpsol --cpxlp .ILP-FILE -o .ILP-FILE-OUT 1>/dev/null");
    else system("glpsol --cpxlp .ILP-FILE -o .ILP-FILE-OUT"); 

    if((fp=fopen(".ILP-FILE-OUT","r"))==0)
     {
        printf("Cannot open file to read the Integer Programming Output\n");
        exit(0);
     }


   fprintf(fr,"\n\n\n");
   fprintf(fr,"USER DEFINED PARAMETERS:\n");
   fprintf(fr,"PRIMER LENGTH                      : %ld\n", (long)k);
   fprintf(fr,"PRODUCT LENGTH                     : %ld\n", (long)L);
   fprintf(fr,"MIN PROXIMITY TO SNP               : %ld\n",(long)prox);
   fprintf(fr,"NUMBER OF TARGETS                  : %ld\n", m);
   fprintf(fr,"MELTING TEMP RANGE                 : ( %f , %f )\n", min_candid_temp, max_candid_temp);
   fprintf(fr,"GC CONTENT RANGE                   : ( %f , %f )\n", min_candid_gc,max_candid_gc);
   fprintf(fr,"PRIMER SALT CONC                   : %f\n", candid_salt_conc);
   fprintf(fr,"PRIMER DNA CONC                    : %f\n\n\n", candid_dna_conc);





/* list out the primers   */
  for(i=0;i<k-6; i++)
    fprintf(fr," ");
  fprintf(fr,"PRIMER(5'-3')       MELT  TEMP            GC\n");



  i=0;
#define  MAX_BUF  200
  int count, state, value=-1;
  char primerstr[50], buf[MAX_BUF];
  long primerno=-1;
  long tot_primers=0;
  
  while(!feof(fp)) {
     fgets(buf, MAX_BUF,fp);
     
     i=0; state=0; count = 0; value =-1;
     while(buf[i] !='\n' &&  buf[i]!='\0')
       {
         if(state==0 && buf[i]=='x')
           state++;
         else if(state==1 && buf[i]=='(')
           {
             state++;
           }
         else if(state==1 && buf[i]!='(')
           {
              state= 10;
           }
         else if(state ==2 && isdigit(buf[i]))
            primerstr[count++] = buf[i];
         else if(state==2 && buf[i] == ')')
           {
             state++;
             primerstr[count]='\0';
             primerno=atol(primerstr);
           }
         else if(state==3 && isdigit(buf[i]))
           {
             state++;
             if(buf[i]=='1') value =1; 
             else if(buf[i]=='0') value =0;
             else 
               {
                 perror("Error reading ILP-FILE-OUT\n");
                 exit(0);
               }
           }

         i=i+1;
     } /*end innter while */

      if(state ==4 &&  value ==1)
        {

      for( j = 0; j < k; j++ )
       {
           Seq[j]=toupper(code[(int)pair[primer[primerno].pair].str[primer[primerno].str][primer[primerno].pos+j]]);
           Sequen[j]=code[(int)pair[primer[primerno].pair].str[primer[primerno].str][primer[primerno].pos+j]];
       
      }                  

         Seq[k] ='\0';
         Sequen[k] ='\0';
         temperature=seqtm(Seq, candid_salt_conc, candid_dna_conc, MAX_NN_TM_LENGTH, strlen(Seq));
         gc_content=gc_and_n_content(Seq, strlen(Seq));
         fprintf(fr,"   %s         %6.2lf            %6.2lf\n",Sequen, temperature, gc_content);
         tot_primers++;
      }

     c = fgetc(fp);
     if(c == EOF)
       break;
     else
       fseek(fp,-1,SEEK_CUR);
  } /* end of reading ILP-FILE-OUT file */


  fprintf(fr,"\n\nNUMBER OF PRIMERS SELECTED         : %ld\n\n",tot_primers) ;


  fclose(fp);
  fclose(fr);

  if(Seq!=NULL) free(Seq);
  if(Sequen!=NULL) free(Sequen);
  return 0;
}


